@@ -650,11 +650,6 @@ redirect_windows (MetaCompositor *compositor,
   guint        n_retries;
   guint        max_retries;
 
-  /* If we're running with wayland, connected to a headless xwayland
-   * server then all the windows are implicitly redirected offscreen
-   * already and it would generate an error to try and explicitly
-   * redirect them via XCompositeRedirectSubwindows() */
-
   if (meta_get_replace_current_wm ())
     max_retries = 5;
   else
@@ -792,8 +787,6 @@ meta_compositor_manage_screen (MetaCompositor *compositor,
   clutter_actor_add_child (info->stage, info->window_group);
   clutter_actor_add_child (info->stage, info->top_window_group);
 
-  info->plugin_mgr = meta_plugin_manager_new (screen);
-
   if (meta_is_wayland_compositor ())
     {
       /* NB: When running as a wayland compositor we don't need an X
@@ -803,13 +796,6 @@ meta_compositor_manage_screen (MetaCompositor *compositor,
     }
   else
     {
-      /*
-       * Delay the creation of the overlay window as long as we can, to avoid
-       * blanking out the screen. This means that during the plugin loading, the
-       * overlay window is not accessible; if the plugin needs to access it
-       * directly, it should hook into the "show" signal on stage, and do
-       * its stuff there.
-       */
       info->output = get_output_window (screen);
       XReparentWindow (xdisplay, xwin, info->output, 0, 0);
 
@@ -830,6 +816,8 @@ meta_compositor_manage_screen (MetaCompositor *compositor,
           info->pending_input_region = None;
         }
 
+      info->plugin_mgr = meta_plugin_manager_new (screen);
+
       /* Map overlay window before redirecting windows offscreen so we catch their
        * contents until we show the stage.
        */
@@ -876,7 +864,7 @@ meta_shape_cow_for_window (MetaScreen *screen,
       int width, height;
       MetaRectangle rect;
 
-      meta_window_get_outer_rect (metaWindow, &rect);
+      meta_window_get_frame_rect (metaWindow, &rect);
 
       window_bounds.x = rect.x;
       window_bounds.y = rect.y;
