@@ -49,6 +49,7 @@
 
 #include "meta-wayland-pointer.h"
 #include "meta-wayland-private.h"
+#include "xdg-shell-server-protocol.h"
 
 #include <string.h>
 
@@ -352,6 +353,11 @@ meta_wayland_pointer_set_focus (MetaWaylandPointer *pointer,
         }
 
       meta_wayland_pointer_get_relative_coordinates (pointer, surface, &sx, &sy);
+      meta_window_handle_enter (surface->window,
+                                /* XXX -- can we reliably get a timestamp for setting focus? */
+                                clutter_get_current_event_time (),
+                                wl_fixed_to_int (pointer->x),
+                                wl_fixed_to_int (pointer->y));
       wl_pointer_send_enter (resource, serial, surface->resource, sx, sy);
       wl_resource_add_destroy_listener (resource, &pointer->focus_listener);
       pointer->focus_serial = serial;
@@ -532,9 +538,14 @@ meta_wayland_pointer_end_popup_grab (MetaWaylandPointer *pointer)
 
   wl_list_for_each_safe (popup, tmp, &popup_grab->all_popups, link)
     {
-      MetaWaylandSurfaceExtension *shell_surface = popup->surface->shell_surface;
+      MetaWaylandSurfaceExtension *shell_surface = popup->surface->xdg_surface;
+      struct wl_client *client = wl_resource_get_client (shell_surface->resource);
+      struct wl_display *display = wl_client_get_display (client);
+      uint32_t serial;
 
-      wl_shell_surface_send_popup_done (shell_surface->resource);
+      serial = wl_display_next_serial (display);
+
+      xdg_popup_send_popup_done (shell_surface->resource, serial);
       wl_list_remove (&popup->surface_destroy_listener.link);
       wl_list_remove (&popup->link);
       g_slice_free (MetaWaylandPopup, popup);
