@@ -51,14 +51,7 @@ static void
 pointer_unmap_sprite (MetaWaylandSeat *seat)
 {
   if (seat->cursor_tracker)
-    {
-      meta_cursor_tracker_set_buffer (seat->cursor_tracker,
-				      NULL, 0, 0);
-
-      if (seat->current_stage)
-	meta_cursor_tracker_queue_redraw (seat->cursor_tracker,
-					  CLUTTER_ACTOR (seat->current_stage));
-    }
+    meta_cursor_tracker_set_window_cursor (seat->cursor_tracker, NULL, 0, 0);
 
   if (seat->sprite)
     {
@@ -76,14 +69,10 @@ meta_wayland_seat_update_sprite (MetaWaylandSeat *seat)
     return;
 
   buffer = seat->sprite->buffer_ref.buffer->resource;
-  meta_cursor_tracker_set_buffer (seat->cursor_tracker,
-				  buffer,
-				  seat->hotspot_x,
-				  seat->hotspot_y);
-
-  if (seat->current_stage)
-    meta_cursor_tracker_queue_redraw (seat->cursor_tracker,
-				      CLUTTER_ACTOR (seat->current_stage));
+  meta_cursor_tracker_set_window_cursor (seat->cursor_tracker,
+                                         buffer,
+                                         seat->hotspot_x,
+                                         seat->hotspot_y);
 }
 
 static void
@@ -257,11 +246,6 @@ notify_motion (MetaWaylandSeat    *seat,
                const ClutterEvent *event)
 {
   MetaWaylandPointer *pointer = &seat->pointer;
-  float x, y;
-
-  clutter_event_get_coords (event, &x, &y);
-  pointer->x = wl_fixed_from_double (x);
-  pointer->y = wl_fixed_from_double (y);
 
   meta_wayland_seat_repick (seat, event);
 
@@ -280,34 +264,22 @@ handle_button_event (MetaWaylandSeat    *seat,
                      const ClutterEvent *event)
 {
   MetaWaylandPointer *pointer = &seat->pointer;
-  gboolean state = event->type == CLUTTER_BUTTON_PRESS;
-  uint32_t button;
-  MetaWaylandSurface *surface;
+  gboolean implicit_grab;
 
   notify_motion (seat, event);
 
-  if (state && pointer->button_count == 1)
+  implicit_grab = (event->type == CLUTTER_BUTTON_PRESS) && (pointer->button_count == 1);
+  if (implicit_grab)
     {
-      button = clutter_event_get_button (event);
-      pointer->grab_button = button;
+      pointer->grab_button = clutter_event_get_button (event);
       pointer->grab_time = clutter_event_get_time (event);
       pointer->grab_x = pointer->x;
       pointer->grab_y = pointer->y;
-
-      /* FIXME: synth a XI2 event and handle in display.c */
-      surface = pointer->current;
-      if (button == CLUTTER_BUTTON_PRIMARY &&
-	  surface &&
-	  surface->window &&
-	  surface->window->client_type == META_WINDOW_CLIENT_TYPE_WAYLAND)
-	{
-	  meta_window_raise (surface->window);
-	}
     }
 
   pointer->grab->interface->button (pointer->grab, event);
 
-  if (pointer->button_count == 1)
+  if (implicit_grab)
     pointer->grab_serial = wl_display_get_serial (seat->display);
 }
 
@@ -375,12 +347,33 @@ count_buttons (const ClutterEvent *event)
   return count;
 }
 
+void
+meta_wayland_seat_update_pointer (MetaWaylandSeat    *seat,
+                                  const ClutterEvent *event)
+{
+  float x, y;
+
+  clutter_event_get_coords (event, &x, &y);
+  seat->pointer.x = wl_fixed_from_double (x);
+  seat->pointer.y = wl_fixed_from_double (y);
+
+  seat->pointer.button_count = count_buttons (event);
+
+  if (seat->cursor_tracker)
+    {
+      meta_cursor_tracker_update_position (seat->cursor_tracker,
+					   wl_fixed_to_int (seat->pointer.x),
+					   wl_fixed_to_int (seat->pointer.y));
+
+      if (seat->pointer.current == NULL)
+	meta_cursor_tracker_unset_window_cursor (seat->cursor_tracker);
+    }
+}
+
 gboolean
 meta_wayland_seat_handle_event (MetaWaylandSeat *seat,
                                 const ClutterEvent *event)
 {
-  seat->pointer.button_count = count_buttons (event);
-
   switch (event->type)
     {
     case CLUTTER_MOTION:
@@ -408,20 +401,6 @@ meta_wayland_seat_handle_event (MetaWaylandSeat *seat,
   return FALSE;
 }
 
-static void
-update_pointer_position_for_actor (MetaWaylandPointer *pointer,
-                                   ClutterActor *actor)
-{
-  float ax, ay;
-
-  clutter_actor_transform_stage_point (actor,
-                                       wl_fixed_to_double (pointer->x),
-                                       wl_fixed_to_double (pointer->y),
-                                       &ax, &ay);
-  pointer->current_x = wl_fixed_from_double (ax);
-  pointer->current_y = wl_fixed_from_double (ay);
-}
-
 /* The actor argument can be NULL in which case a Clutter pick will be
    performed to determine the right actor. An actor should only be
    passed if the repick is being performed due to an event in which
@@ -458,18 +437,8 @@ meta_wayland_seat_repick (MetaWaylandSeat    *seat,
       MetaWindow *window =
         meta_window_actor_get_meta_window (META_WINDOW_ACTOR (actor));
 
-      update_pointer_position_for_actor (pointer, actor);
-
       surface = window->surface;
     }
-  else if (META_IS_SHAPED_TEXTURE (actor))
-    {
-      MetaShapedTexture *shaped_texture = META_SHAPED_TEXTURE (actor);
-
-      update_pointer_position_for_actor (pointer, actor);
-
-      surface = meta_shaped_texture_get_wayland_surface (shaped_texture);
-    }
 
   pointer->current = surface;
   if (surface != pointer->focus)
