@@ -39,6 +39,7 @@
 
 #include <wayland-server.h>
 #include "gtk-shell-server-protocol.h"
+#include "xdg-shell-server-protocol.h"
 
 #include "meta-wayland-private.h"
 #include "meta-xwayland-private.h"
@@ -58,9 +59,6 @@
 #include "meta-weston-launch.h"
 #include "monitor-private.h"
 
-static void ensure_initial_state (MetaWaylandSurface *surface);
-static void free_initial_state   (MetaWaylandSurfaceInitialState *surface);
-
 static void
 surface_process_damage (MetaWaylandSurface *surface,
                         cairo_region_t *region)
@@ -199,14 +197,17 @@ meta_wayland_surface_set_opaque_region (struct wl_client *client,
                                         struct wl_resource *region_resource)
 {
   MetaWaylandSurface *surface = wl_resource_get_user_data (surface_resource);
-  MetaWaylandRegion *region = wl_resource_get_user_data (region_resource);
 
   /* X11 unmanaged window */
   if (!surface)
     return;
 
   g_clear_pointer (&surface->pending.opaque_region, cairo_region_destroy);
-  surface->pending.opaque_region = cairo_region_copy (region->region);
+  if (region_resource)
+    {
+      MetaWaylandRegion *region = wl_resource_get_user_data (region_resource);
+      surface->pending.opaque_region = cairo_region_copy (region->region);
+    }
 }
 
 static void
@@ -215,14 +216,17 @@ meta_wayland_surface_set_input_region (struct wl_client *client,
                                        struct wl_resource *region_resource)
 {
   MetaWaylandSurface *surface = wl_resource_get_user_data (surface_resource);
-  MetaWaylandRegion *region = wl_resource_get_user_data (region_resource);
 
   /* X11 unmanaged window */
   if (!surface)
     return;
 
   g_clear_pointer (&surface->pending.input_region, cairo_region_destroy);
-  surface->pending.input_region = cairo_region_copy (region->region);
+  if (region_resource)
+    {
+      MetaWaylandRegion *region = wl_resource_get_user_data (region_resource);
+      surface->pending.input_region = cairo_region_copy (region->region);
+    }
 }
 
 static void
@@ -232,6 +236,52 @@ empty_region (cairo_region_t *region)
   cairo_region_intersect_rectangle (region, &rectangle);
 }
 
+static gboolean
+surface_wants_window (MetaWaylandSurface *surface)
+{
+  return (surface->xdg_surface != NULL);
+}
+
+static void
+surface_ensure_window (MetaWaylandSurface *surface)
+{
+  MetaDisplay *display = meta_get_display ();
+  int width, height;
+
+  if (surface->window)
+    return;
+
+  if (!surface_wants_window (surface))
+    return;
+
+  width = surface->buffer_ref.buffer->width;
+  height = surface->buffer_ref.buffer->height;
+
+  surface->window = meta_window_new_for_wayland (display, width, height, surface);
+}
+
+static void
+ensure_buffer_texture (MetaWaylandBuffer *buffer)
+{
+  CoglContext *ctx = clutter_backend_get_cogl_context (clutter_get_default_backend ());
+  CoglError *catch_error = NULL;
+  CoglTexture *texture;
+
+  texture = COGL_TEXTURE (cogl_wayland_texture_2d_new_from_buffer (ctx,
+                                                                   buffer->resource,
+                                                                   &catch_error));
+  if (!texture)
+    {
+      cogl_error_free (catch_error);
+      meta_warning ("Could not import pending buffer, ignoring commit\n");
+      return;
+    }
+
+  buffer->texture = texture;
+  buffer->width = cogl_texture_get_width (texture);
+  buffer->height = cogl_texture_get_height (texture);
+}
+
 static void
 meta_wayland_surface_commit (struct wl_client *client,
                              struct wl_resource *resource)
@@ -250,51 +300,18 @@ meta_wayland_surface_commit (struct wl_client *client,
       surface->buffer_ref.buffer != surface->pending.buffer)
     {
       MetaWaylandBuffer *buffer = surface->pending.buffer;
-      CoglContext *ctx =
-        clutter_backend_get_cogl_context (clutter_get_default_backend ());
-      CoglError *catch_error = NULL;
-      CoglTexture *texture =
-        COGL_TEXTURE (cogl_wayland_texture_2d_new_from_buffer (ctx,
-                                                               buffer->resource,
-                                                               &catch_error));
-      if (!texture)
+
+      if (buffer)
         {
-          cogl_error_free (catch_error);
-	  meta_warning ("Could not import pending buffer, ignoring commit\n");
-	  return;
+          /* Note: we set this before informing any window-actor since the
+           * window actor will expect to find the new buffer within the
+           * surface. */
+          ensure_buffer_texture (buffer);
+          meta_wayland_buffer_reference (&surface->buffer_ref, buffer);
         }
-      else
-        {
-	  buffer->texture = texture;
-          buffer->width = cogl_texture_get_width (texture);
-          buffer->height = cogl_texture_get_height (texture);
-        }
-
-      /* Note: we set this before informing any window-actor since the
-       * window actor will expect to find the new buffer within the
-       * surface. */
-      meta_wayland_buffer_reference (&surface->buffer_ref,
-                                     surface->pending.buffer);
     }
 
-  if (!surface->buffer_ref.buffer)
-    {
-      meta_warning ("Commit without a buffer? Ignoring\n");
-      return;
-    }
-
-  if (surface->initial_state)
-    {
-      MetaDisplay *display = meta_get_display ();
-      int width;
-      int height;
-
-      width = surface->buffer_ref.buffer->width;
-      height = surface->buffer_ref.buffer->height;
-
-      /* This will free and clear initial_state */
-      surface->window = meta_window_new_for_wayland (display, width, height, surface);
-    }
+  surface_ensure_window (surface);
 
   if (surface == compositor->seat->sprite)
     meta_wayland_seat_update_sprite (compositor->seat);
@@ -329,6 +346,27 @@ meta_wayland_surface_commit (struct wl_client *client,
       meta_window_set_opaque_region (surface->window, surface->pending.opaque_region);
       meta_window_set_input_region (surface->window, surface->pending.input_region);
       surface_process_damage (surface, surface->pending.damage);
+
+      meta_window_set_title (surface->window, surface->pending.title);
+      g_clear_pointer (&surface->pending.title, g_free);
+
+      meta_window_set_wm_class (surface->window, surface->pending.app_id, surface->pending.app_id);
+      g_clear_pointer (&surface->pending.app_id, g_free);
+
+      meta_window_set_gtk_dbus_properties (surface->window,
+                                           surface->pending.gtk_application_id,
+                                           surface->pending.gtk_unique_bus_name,
+                                           surface->pending.gtk_app_menu_path,
+                                           surface->pending.gtk_menubar_path,
+                                           surface->pending.gtk_application_object_path,
+                                           surface->pending.gtk_window_object_path);
+
+      g_clear_pointer (&surface->pending.gtk_application_id, g_free);
+      g_clear_pointer (&surface->pending.gtk_unique_bus_name, g_free);
+      g_clear_pointer (&surface->pending.gtk_app_menu_path, g_free);
+      g_clear_pointer (&surface->pending.gtk_menubar_path, g_free);
+      g_clear_pointer (&surface->pending.gtk_application_object_path, g_free);
+      g_clear_pointer (&surface->pending.gtk_window_object_path, g_free);
     }
 
   if (surface->pending.buffer)
@@ -362,7 +400,8 @@ meta_wayland_surface_set_buffer_scale (struct wl_client *client,
                                        struct wl_resource *resource,
                                        int scale)
 {
-  g_warning ("TODO: support set_buffer_scale request");
+  if (scale != 1)
+    g_warning ("TODO: support set_buffer_scale request");
 }
 
 const struct wl_surface_interface meta_wayland_surface_interface = {
@@ -383,9 +422,6 @@ meta_wayland_surface_free (MetaWaylandSurface *surface)
   MetaWaylandCompositor *compositor = surface->compositor;
   MetaWaylandFrameCallback *cb, *next;
 
-  if (surface->initial_state)
-    free_initial_state (surface->initial_state);
-
   compositor->surfaces = g_list_remove (compositor->surfaces, surface);
 
   meta_wayland_buffer_reference (&surface->buffer_ref, NULL);
@@ -409,9 +445,6 @@ meta_wayland_surface_free (MetaWaylandSurface *surface)
       g_assert (surface != compositor->seat->pointer.focus);
     }
 
- if (compositor->implicit_grab_surface == surface)
-   compositor->implicit_grab_surface = compositor->seat->pointer.current;
-
   if (surface->resource)
     wl_resource_set_user_data (surface->resource, NULL);
   g_slice_free (MetaWaylandSurface, surface);
@@ -486,339 +519,223 @@ meta_wayland_surface_create (MetaWaylandCompositor *compositor,
 }
 
 static void
-shell_surface_pong (struct wl_client *client,
-                    struct wl_resource *resource,
-                    guint32 serial)
+xdg_surface_destroy (struct wl_client *client,
+                     struct wl_resource *resource)
 {
-}
-
-typedef struct
-{
-  MetaWaylandPointerGrab  generic;
-  MetaWaylandSurface     *surface;
-  struct wl_listener      surface_destroy_listener;
-  wl_fixed_t              dx, dy;
-} MetaWaylandMoveGrab;
-
-static void
-end_move_grab (MetaWaylandMoveGrab *move_grab)
-{
-  if (move_grab->surface)
-    {
-      move_grab->surface = NULL;
-      wl_list_remove (&move_grab->surface_destroy_listener.link);
-    }
-
-  meta_wayland_pointer_end_grab (move_grab->generic.pointer);
-  g_slice_free (MetaWaylandMoveGrab, move_grab);
+  g_warning ("TODO: support xdg_surface.destroy");
 }
 
 static void
-move_grab_lose_surface (struct wl_listener *listener, void *data)
+xdg_surface_set_transient_for (struct wl_client *client,
+                               struct wl_resource *resource,
+                               struct wl_resource *parent)
 {
-  MetaWaylandMoveGrab *move_grab =
-    wl_container_of (listener, move_grab, surface_destroy_listener);
+  MetaWaylandSurfaceExtension *surface_ext = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = surface_ext->surface;
 
-  move_grab->surface = NULL;
-  end_move_grab (move_grab);
+  MetaWaylandSurfaceExtension *parent_ext = wl_resource_get_user_data (parent);
+  MetaWaylandSurface *parent_surf = parent_ext->surface;
+
+  if (surface->window && parent_surf->window)
+    meta_window_set_transient_for (surface->window, parent_surf->window);
 }
 
 static void
-move_grab_focus (MetaWaylandPointerGrab *grab,
-                 MetaWaylandSurface     *surface,
-		 const ClutterEvent     *event)
+xdg_surface_set_title (struct wl_client *client,
+                       struct wl_resource *resource,
+                       const char *title)
 {
+  MetaWaylandSurfaceExtension *extension = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = extension->surface;
+
+  g_clear_pointer (&surface->pending.title, g_free);
+  surface->pending.title = g_strdup (title);
 }
 
 static void
-move_grab_motion (MetaWaylandPointerGrab *grab,
-		  const ClutterEvent     *event)
+xdg_surface_set_app_id (struct wl_client *client,
+                        struct wl_resource *resource,
+                        const char *app_id)
 {
-  MetaWaylandMoveGrab *move_grab = (MetaWaylandMoveGrab *)grab;
-  MetaWaylandPointer *pointer = grab->pointer;
+  MetaWaylandSurfaceExtension *extension = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = extension->surface;
 
-  meta_window_move (move_grab->surface->window,
-                    TRUE,
-                    wl_fixed_to_int (pointer->x + move_grab->dx),
-                    wl_fixed_to_int (pointer->y + move_grab->dy));
+  g_clear_pointer (&surface->pending.app_id, g_free);
+  surface->pending.app_id = g_strdup (app_id);
 }
 
 static void
-move_grab_button (MetaWaylandPointerGrab *grab,
-		  const ClutterEvent     *event)
+xdg_surface_pong (struct wl_client *client,
+                  struct wl_resource *resource,
+                  guint32 serial)
 {
-  MetaWaylandMoveGrab *move_grab = (MetaWaylandMoveGrab *)grab;
-  MetaWaylandPointer *pointer = grab->pointer;
-
-  if (pointer->button_count == 0 &&
-      clutter_event_type (event) == CLUTTER_BUTTON_RELEASE)
-    end_move_grab (move_grab);
 }
 
-static const MetaWaylandPointerGrabInterface move_grab_interface = {
-    move_grab_focus,
-    move_grab_motion,
-    move_grab_button,
-};
+static gboolean
+begin_grab_op_on_surface (MetaWaylandSurface *surface,
+                          MetaWaylandSeat    *seat,
+                          MetaGrabOp          grab_op)
+{
+  MetaWindow *window = surface->window;
+
+  if (!window)
+    return FALSE;
+
+  if (grab_op == META_GRAB_OP_NONE)
+    return FALSE;
+
+  return meta_display_begin_grab_op (window->display,
+                                     window->screen,
+                                     window,
+                                     grab_op,
+                                     TRUE, /* pointer_already_grabbed */
+                                     FALSE, /* frame_action */
+                                     1, /* button. XXX? */
+                                     0, /* modmask */
+                                     meta_display_get_current_time_roundtrip (window->display),
+                                     wl_fixed_to_int (seat->pointer.grab_x),
+                                     wl_fixed_to_int (seat->pointer.grab_y));
+}
 
 static void
-shell_surface_move (struct wl_client *client,
-                    struct wl_resource *resource,
-                    struct wl_resource *seat_resource,
-                    guint32 serial)
+xdg_surface_move (struct wl_client *client,
+                  struct wl_resource *resource,
+                  struct wl_resource *seat_resource,
+                  guint32 serial)
 {
   MetaWaylandSeat *seat = wl_resource_get_user_data (seat_resource);
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-  MetaWindow *window;
-  MetaWaylandMoveGrab *move_grab;
-  MetaRectangle rect;
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
 
   if (seat->pointer.button_count == 0 ||
       seat->pointer.grab_serial != serial ||
-      seat->pointer.focus != shell_surface->surface)
+      seat->pointer.focus != xdg_surface->surface)
     return;
 
-  window = shell_surface->surface->window;
-  if (!window)
-    return;
-
-  /* small hack, this should be handled by window->shaken_loose when we
-     move everything to display.c */
-  if (window->fullscreen)
-    meta_window_unmake_fullscreen (window);
-
-  move_grab = g_slice_new (MetaWaylandMoveGrab);
-
-  meta_window_get_input_rect (shell_surface->surface->window,
-                              &rect);
-
-  move_grab->generic.interface = &move_grab_interface;
-  move_grab->generic.pointer = &seat->pointer;
-
-  move_grab->surface = shell_surface->surface;
-  move_grab->surface_destroy_listener.notify = move_grab_lose_surface;
-  wl_resource_add_destroy_listener (shell_surface->surface->resource,
-				    &move_grab->surface_destroy_listener);
-
-  move_grab->dx = wl_fixed_from_int (rect.x) - seat->pointer.grab_x;
-  move_grab->dy = wl_fixed_from_int (rect.y) - seat->pointer.grab_y;
-
-  meta_wayland_pointer_start_grab (&seat->pointer, (MetaWaylandPointerGrab*)move_grab);
-
-  /* TODO: send_grab_cursor (cursor); */
-
-  /* XXX: In Weston there is a desktop shell protocol which has
-   * a set_grab_surface request that's used to specify the surface
-   * that's focused here.
-   *
-   * TODO: understand why.
-   *
-   * XXX: For now we just focus the surface directly associated with
-   * the grab.
-   */
-  meta_wayland_pointer_set_focus (&seat->pointer, shell_surface->surface);
+  begin_grab_op_on_surface (xdg_surface->surface, seat, META_GRAB_OP_MOVING);
 }
 
-static void
-shell_surface_resize (struct wl_client *client,
-                      struct wl_resource *resource,
-                      struct wl_resource *seat,
-                      guint32 serial,
-                      guint32 edges)
+static MetaGrabOp
+grab_op_for_edge (int edge)
 {
-  g_warning ("TODO: support shell_surface_resize request");
+  switch (edge)
+    {
+    case XDG_SURFACE_RESIZE_EDGE_TOP_LEFT:
+      return META_GRAB_OP_RESIZING_NW;
+    case XDG_SURFACE_RESIZE_EDGE_TOP:
+      return META_GRAB_OP_RESIZING_N;
+    case XDG_SURFACE_RESIZE_EDGE_TOP_RIGHT:
+      return META_GRAB_OP_RESIZING_NE;
+    case XDG_SURFACE_RESIZE_EDGE_RIGHT:
+      return META_GRAB_OP_RESIZING_E;
+    case XDG_SURFACE_RESIZE_EDGE_BOTTOM_RIGHT:
+      return META_GRAB_OP_RESIZING_SE;
+    case XDG_SURFACE_RESIZE_EDGE_BOTTOM:
+      return META_GRAB_OP_RESIZING_S;
+    case XDG_SURFACE_RESIZE_EDGE_BOTTOM_LEFT:
+      return META_GRAB_OP_RESIZING_SW;
+    case XDG_SURFACE_RESIZE_EDGE_LEFT:
+      return META_GRAB_OP_RESIZING_W;
+    default:
+      g_warning ("invalid edge: %d", edge);
+      return META_GRAB_OP_NONE;
+    }
 }
 
 static void
-shell_surface_set_toplevel (struct wl_client *client,
+xdg_surface_resize (struct wl_client *client,
+                    struct wl_resource *resource,
+                    struct wl_resource *seat_resource,
+                    guint32 serial,
+                    guint32 edges)
+{
+  MetaWaylandSeat *seat = wl_resource_get_user_data (seat_resource);
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
+
+  if (seat->pointer.button_count == 0 ||
+      seat->pointer.grab_serial != serial ||
+      seat->pointer.focus != xdg_surface->surface)
+    return;
+
+  begin_grab_op_on_surface (xdg_surface->surface, seat, grab_op_for_edge (edges));
+}
+
+static void
+xdg_surface_set_output (struct wl_client *client,
+                        struct wl_resource *resource,
+                        struct wl_resource *output)
+{
+  g_warning ("TODO: support xdg_surface.set_output");
+}
+
+static void
+xdg_surface_set_fullscreen (struct wl_client *client,
                             struct wl_resource *resource)
 {
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = shell_surface->surface;
-  MetaWaylandCompositor *compositor = surface->compositor;
-
-  /* NB: Surfaces from xwayland become managed based on X events. */
-  if (client == compositor->xwayland_client)
-    return;
-
-  if (surface->window)
-    {
-      if (surface->window->fullscreen)
-	meta_window_unmake_fullscreen (surface->window);
-      if (meta_window_get_maximized (surface->window) != 0)
-	meta_window_unmaximize (surface->window, META_MAXIMIZE_HORIZONTAL | META_MAXIMIZE_VERTICAL);
-    }
-  else
-    {
-      ensure_initial_state (surface);
-
-      surface->initial_state->initial_type = META_WAYLAND_SURFACE_TOPLEVEL;
-    }
-}
-
-static void
-shell_surface_set_transient (struct wl_client *client,
-                             struct wl_resource *resource,
-                             struct wl_resource *parent,
-                             int x,
-                             int y,
-                             guint32 flags)
-{
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = shell_surface->surface;
-  MetaWaylandSurface *parent_surface = wl_resource_get_user_data (parent);
-  MetaWaylandCompositor *compositor = surface->compositor;
-
-  /* NB: Surfaces from xwayland become managed based on X events. */
-  if (client == compositor->xwayland_client)
-    return;
-
-  if (surface->window)
-    meta_window_set_transient_for (surface->window, parent_surface->window);
-  else
-    {
-      ensure_initial_state (surface);
-
-      surface->initial_state->initial_type = META_WAYLAND_SURFACE_TOPLEVEL;
-      surface->initial_state->transient_for = parent;
-    }
-}
-
-static void
-shell_surface_set_fullscreen (struct wl_client *client,
-                              struct wl_resource *resource,
-                              guint32 method,
-                              guint32 framerate,
-                              struct wl_resource *output)
-{
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = shell_surface->surface;
-  MetaWaylandCompositor *compositor = surface->compositor;
-
-  /* NB: Surfaces from xwayland become managed based on X events. */
-  if (client == compositor->xwayland_client)
-    return;
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = xdg_surface->surface;
 
   if (surface->window)
     meta_window_make_fullscreen (surface->window);
-  else
-    {
-      ensure_initial_state (surface);
-
-      surface->initial_state->initial_type = META_WAYLAND_SURFACE_FULLSCREEN;
-    }
 }
 
 static void
-shell_surface_set_popup (struct wl_client *client,
-                         struct wl_resource *resource,
-                         struct wl_resource *seat_resource,
-                         guint32 serial,
-                         struct wl_resource *parent,
-                         gint32 x,
-                         gint32 y,
-                         guint32 flags)
+xdg_surface_unset_fullscreen (struct wl_client *client,
+                              struct wl_resource *resource)
 {
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = shell_surface->surface;
-  MetaWaylandCompositor *compositor = surface->compositor;
-  MetaWaylandSeat *seat = compositor->seat;
-
-  if (serial < seat->pointer.click_serial)
-    {
-      /* stale request */
-      return;
-    }
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = xdg_surface->surface;
 
   if (surface->window)
-    {
-      meta_warning ("Client set_popup() on an already visible window, this is not supported\n");
-    }
-  else
-    {
-      ensure_initial_state (surface);
-
-      surface->initial_state->initial_type = META_WAYLAND_SURFACE_POPUP;
-      surface->initial_state->transient_for = parent;
-      surface->initial_state->x = x;
-      surface->initial_state->y = y;
-    }
-
+    meta_window_unmake_fullscreen (surface->window);
 }
 
 static void
-shell_surface_set_maximized (struct wl_client *client,
-                             struct wl_resource *resource,
-                             struct wl_resource *output)
+xdg_surface_set_maximized (struct wl_client *client,
+                           struct wl_resource *resource)
 {
-  MetaWaylandSurfaceExtension *shell_surface = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = shell_surface->surface;
-  MetaWaylandCompositor *compositor = surface->compositor;
-
-  /* NB: Surfaces from xwayland become managed based on X events. */
-  if (client == compositor->xwayland_client)
-    return;
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = xdg_surface->surface;
 
   if (surface->window)
     meta_window_maximize (surface->window, META_MAXIMIZE_HORIZONTAL | META_MAXIMIZE_VERTICAL);
-  else
-    {
-      ensure_initial_state (surface);
-
-      surface->initial_state->initial_type = META_WAYLAND_SURFACE_MAXIMIZED;
-    }
 }
 
 static void
-shell_surface_set_title (struct wl_client *client,
-                         struct wl_resource *resource,
-                         const char *title)
+xdg_surface_unset_maximized (struct wl_client *client,
+                             struct wl_resource *resource)
 {
-  MetaWaylandSurfaceExtension *extension = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = extension->surface;
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = xdg_surface->surface;
 
   if (surface->window)
-    meta_window_set_title (surface->window, title);
-  else
-    {
-      ensure_initial_state (surface);
-
-      g_free (surface->initial_state->title);
-      surface->initial_state->title = g_strdup (title);
-    }
+    meta_window_unmaximize (surface->window, META_MAXIMIZE_HORIZONTAL | META_MAXIMIZE_VERTICAL);
 }
 
 static void
-shell_surface_set_class (struct wl_client *client,
-                         struct wl_resource *resource,
-                         const char *class_)
+xdg_surface_set_minimized (struct wl_client *client,
+                           struct wl_resource *resource)
 {
-  MetaWaylandSurfaceExtension *extension = wl_resource_get_user_data (resource);
-  MetaWaylandSurface *surface = extension->surface;
+  MetaWaylandSurfaceExtension *xdg_surface = wl_resource_get_user_data (resource);
+  MetaWaylandSurface *surface = xdg_surface->surface;
 
   if (surface->window)
-    meta_window_set_wm_class (surface->window, class_, class_);
-  else
-    {
-      ensure_initial_state (surface);
-
-      g_free (surface->initial_state->wm_class);
-      surface->initial_state->wm_class = g_strdup (class_);
-    }
+    meta_window_minimize (surface->window);
 }
 
-static const struct wl_shell_surface_interface meta_wayland_shell_surface_interface =
-{
-  shell_surface_pong,
-  shell_surface_move,
-  shell_surface_resize,
-  shell_surface_set_toplevel,
-  shell_surface_set_transient,
-  shell_surface_set_fullscreen,
-  shell_surface_set_popup,
-  shell_surface_set_maximized,
-  shell_surface_set_title,
-  shell_surface_set_class
+static const struct xdg_surface_interface meta_wayland_xdg_surface_interface = {
+  xdg_surface_destroy,
+  xdg_surface_set_transient_for,
+  xdg_surface_set_title,
+  xdg_surface_set_app_id,
+  xdg_surface_pong,
+  xdg_surface_move,
+  xdg_surface_resize,
+  xdg_surface_set_output,
+  xdg_surface_set_fullscreen,
+  xdg_surface_unset_fullscreen,
+  xdg_surface_set_maximized,
+  xdg_surface_unset_maximized,
+  xdg_surface_set_minimized,
 };
 
 static void
@@ -839,9 +756,7 @@ destroy_surface_extension (struct wl_resource *resource)
 
   /* In case cleaning up a dead client destroys extension first */
   if (extension->surface)
-    {
-      wl_list_remove (&extension->surface_destroy_listener.link);
-    }
+    wl_list_remove (&extension->surface_destroy_listener.link);
 
   g_free (extension);
 }
@@ -873,43 +788,68 @@ create_surface_extension (struct wl_client          *client,
 }
 
 static void
-get_shell_surface (struct wl_client *client,
-                   struct wl_resource *resource,
-                   guint32 id,
-                   struct wl_resource *surface_resource)
+use_unstable_version (struct wl_client *client,
+                      struct wl_resource *resource,
+                      int32_t version)
+{
+  if (version != META_XDG_SHELL_VERSION)
+    g_warning ("Bad xdg_shell version: %d", version);
+}
+
+static void
+get_xdg_surface (struct wl_client *client,
+                 struct wl_resource *resource,
+                 guint32 id,
+                 struct wl_resource *surface_resource)
 {
   MetaWaylandSurface *surface = wl_resource_get_user_data (surface_resource);
 
-  if (surface->shell_surface)
+  if (surface->xdg_surface)
     {
       wl_resource_post_error (surface_resource,
                               WL_DISPLAY_ERROR_INVALID_OBJECT,
-                              "wl_shell::get_shell_surface already requested");
+                              "xdg_shell::get_xdg_surface already requested");
       return;
     }
 
-  surface->shell_surface = create_surface_extension (client, resource, id,
-                                                     META_WL_SHELL_SURFACE_VERSION, surface,
-                                                     &wl_shell_surface_interface,
-                                                     &meta_wayland_shell_surface_interface);
+  surface->xdg_surface = create_surface_extension (client, resource, id,
+                                                   META_XDG_SURFACE_VERSION, surface,
+                                                   &xdg_surface_interface,
+                                                   &meta_wayland_xdg_surface_interface);
 }
 
-static const struct wl_shell_interface meta_wayland_shell_interface =
+static void
+get_xdg_popup (struct wl_client *client,
+               struct wl_resource *resource,
+               uint32_t id,
+               struct wl_resource *surface,
+               struct wl_resource *parent,
+               struct wl_resource *seat,
+               uint32_t serial,
+               int32_t x,
+               int32_t y,
+               uint32_t flags)
 {
-  get_shell_surface
+  g_warning ("TODO: support xdg_shell.get_xdg_popup");
+}
+
+static const struct xdg_shell_interface meta_wayland_xdg_shell_interface = {
+  use_unstable_version,
+  get_xdg_surface,
+  get_xdg_popup,
 };
 
 static void
-bind_shell (struct wl_client *client,
-            void *data,
-            guint32 version,
-            guint32 id)
+bind_xdg_shell (struct wl_client *client,
+                void *data,
+                guint32 version,
+                guint32 id)
 {
   struct wl_resource *resource;
 
-  resource = wl_resource_create (client, &wl_shell_interface,
-				 MIN (META_WL_SHELL_VERSION, version), id);
-  wl_resource_set_implementation (resource, &meta_wayland_shell_interface, data, NULL);
+  resource = wl_resource_create (client, &xdg_shell_interface,
+				 MIN (META_XDG_SHELL_VERSION, version), id);
+  wl_resource_set_implementation (resource, &meta_wayland_xdg_shell_interface, data, NULL);
 }
 
 static void
@@ -933,45 +873,21 @@ set_dbus_properties (struct wl_client   *client,
       return;
     }
 
-  if (surface->window)
-    {
-      meta_window_set_gtk_dbus_properties (surface->window,
-					   application_id,
-					   unique_bus_name,
-					   app_menu_path,
-					   menubar_path,
-					   application_object_path,
-					   window_object_path);
-    }
-  else
-    {
-      MetaWaylandSurfaceInitialState *initial;
-
-      ensure_initial_state (surface);
-      initial = surface->initial_state;
-
-      g_free (initial->gtk_application_id);
-      initial->gtk_application_id = g_strdup (application_id);
-
-      g_free (initial->gtk_unique_bus_name);
-      initial->gtk_unique_bus_name = g_strdup (unique_bus_name);
-
-      g_free (initial->gtk_app_menu_path);
-      initial->gtk_app_menu_path = g_strdup (app_menu_path);
-
-      g_free (initial->gtk_menubar_path);
-      initial->gtk_menubar_path = g_strdup (menubar_path);
-
-      g_free (initial->gtk_application_object_path);
-      initial->gtk_application_object_path = g_strdup (application_object_path);
-
-      g_free (initial->gtk_window_object_path);
-      initial->gtk_window_object_path = g_strdup (window_object_path);
-    }
+  g_clear_pointer (&surface->pending.gtk_application_id, g_free);
+  surface->pending.gtk_application_id = g_strdup (application_id);
+  g_clear_pointer (&surface->pending.gtk_unique_bus_name, g_free);
+  surface->pending.gtk_unique_bus_name = g_strdup (unique_bus_name);
+  g_clear_pointer (&surface->pending.gtk_app_menu_path, g_free);
+  surface->pending.gtk_app_menu_path = g_strdup (app_menu_path);
+  g_clear_pointer (&surface->pending.gtk_menubar_path, g_free);
+  surface->pending.gtk_menubar_path = g_strdup (menubar_path);
+  g_clear_pointer (&surface->pending.gtk_application_object_path, g_free);
+  surface->pending.gtk_application_object_path = g_strdup (application_object_path);
+  g_clear_pointer (&surface->pending.gtk_window_object_path, g_free);
+  surface->pending.gtk_window_object_path = g_strdup (window_object_path);
 }
 
-static const struct gtk_surface_interface meta_wayland_gtk_surface_interface =
-{
+static const struct gtk_surface_interface meta_wayland_gtk_surface_interface = {
   set_dbus_properties
 };
 
@@ -987,7 +903,7 @@ get_gtk_surface (struct wl_client *client,
     {
       wl_resource_post_error (surface_resource,
                               WL_DISPLAY_ERROR_INVALID_OBJECT,
-                              "wl_shell::get_gtk_surface already requested");
+                              "gtk_shell::get_gtk_surface already requested");
       return;
     }
 
@@ -997,8 +913,7 @@ get_gtk_surface (struct wl_client *client,
                                                    &meta_wayland_gtk_surface_interface);
 }
 
-static const struct gtk_shell_interface meta_wayland_gtk_shell_interface =
-{
+static const struct gtk_shell_interface meta_wayland_gtk_shell_interface = {
   get_gtk_surface
 };
 
@@ -1022,10 +937,10 @@ void
 meta_wayland_init_shell (MetaWaylandCompositor *compositor)
 {
   if (wl_global_create (compositor->wayland_display,
-			&wl_shell_interface,
-			META_WL_SHELL_VERSION,
-			compositor, bind_shell) == NULL)
-    g_error ("Failed to register a global shell object");
+			&xdg_shell_interface,
+			META_XDG_SHELL_VERSION,
+			compositor, bind_xdg_shell) == NULL)
+    g_error ("Failed to register a global xdg-shell object");
 
   if (wl_global_create (compositor->wayland_display,
 			&gtk_shell_interface,
@@ -1034,113 +949,28 @@ meta_wayland_init_shell (MetaWaylandCompositor *compositor)
     g_error ("Failed to register a global gtk-shell object");
 }
 
-void
-meta_wayland_surface_set_initial_state (MetaWaylandSurface *surface,
-					MetaWindow         *window)
-{
-  MetaWaylandSurfaceInitialState *initial = surface->initial_state;
-  MetaWaylandCompositor *compositor = surface->compositor;
-  MetaWaylandSeat *seat = compositor->seat;
-
-  if (initial == NULL)
-    return;
-
-  window->type = META_WINDOW_NORMAL;
-
-  /* Note that we poke at the bits directly here, because we're
-     in the middle of meta_window_new_shared() */
-  switch (initial->initial_type)
-    {
-    case META_WAYLAND_SURFACE_TOPLEVEL:
-      break;
-    case META_WAYLAND_SURFACE_FULLSCREEN:
-      window->fullscreen = TRUE;
-      break;
-    case META_WAYLAND_SURFACE_MAXIMIZED:
-      window->maximized_horizontally = window->maximized_vertically = TRUE;
-      break;
-    case META_WAYLAND_SURFACE_POPUP:
-      window->override_redirect = TRUE;
-      window->type = META_WINDOW_DROPDOWN_MENU;
-      window->mapped = TRUE;
-      window->showing_for_first_time = FALSE;
-      window->placed = TRUE;
-      if (!meta_wayland_pointer_start_popup_grab (&seat->pointer, surface))
-	wl_shell_surface_send_popup_done (surface->shell_surface->resource);
-      break;
-    default:
-      g_assert_not_reached ();
-    }
-
-  if (initial->transient_for)
-    {
-      MetaWaylandSurface *parent = wl_resource_get_user_data (initial->transient_for);
-      if (parent && parent->window)
-	{
-	  window->transient_for = g_object_ref (parent->window);
-
-	  if (initial->initial_type == META_WAYLAND_SURFACE_POPUP)
-	    {
-	      window->rect.x = parent->window->rect.x + initial->x;
-	      window->rect.y = parent->window->rect.y + initial->y;
-	    }
-	}
-
-      window->type = META_WINDOW_DIALOG;
-    }
-
-  if (initial->title)
-    meta_window_set_title (window, initial->title);
-
-  if (initial->wm_class)
-    meta_window_set_wm_class (window, initial->wm_class, initial->wm_class);
-
-  meta_window_set_gtk_dbus_properties (window,
-				       initial->gtk_application_id,
-				       initial->gtk_unique_bus_name,
-				       initial->gtk_app_menu_path,
-				       initial->gtk_menubar_path,
-				       initial->gtk_application_object_path,
-				       initial->gtk_window_object_path);
-
-  meta_window_type_changed (window);
-
-  free_initial_state (initial);
-  surface->initial_state = NULL;
-}
-
-static void
-ensure_initial_state (MetaWaylandSurface *surface)
-{
-  if (surface->initial_state)
-    return;
-
-  surface->initial_state = g_slice_new0 (MetaWaylandSurfaceInitialState);
-}
-
-static void
-free_initial_state (MetaWaylandSurfaceInitialState *initial)
-{
-  g_free (initial->title);
-  g_free (initial->wm_class);
-
-  g_free (initial->gtk_application_id);
-  g_free (initial->gtk_unique_bus_name);
-  g_free (initial->gtk_app_menu_path);
-  g_free (initial->gtk_menubar_path);
-  g_free (initial->gtk_application_object_path);
-  g_free (initial->gtk_window_object_path);
-
-  g_slice_free (MetaWaylandSurfaceInitialState, initial);
-}
-
 void
 meta_wayland_surface_configure_notify (MetaWaylandSurface *surface,
 				       int                 new_width,
 				       int                 new_height,
 				       int                 edges)
 {
-  if (surface->shell_surface)
-    wl_shell_surface_send_configure (surface->shell_surface->resource,
-				     edges, new_width, new_height);
+  if (surface->xdg_surface)
+    xdg_surface_send_configure (surface->xdg_surface->resource,
+                                edges, new_width, new_height,
+                                0, 0 /* XXX: support this */);
+}
+
+void
+meta_wayland_surface_focused_set (MetaWaylandSurface *surface)
+{
+  if (surface->xdg_surface)
+    xdg_surface_send_focused_set (surface->xdg_surface->resource);
+}
+
+void
+meta_wayland_surface_focused_unset (MetaWaylandSurface *surface)
+{
+  if (surface->xdg_surface)
+    xdg_surface_send_focused_unset (surface->xdg_surface->resource);
 }
