@@ -64,31 +64,19 @@ typedef struct
 
   /* wl_surface.frame */
   struct wl_list frame_callback_list;
-} MetaWaylandDoubleBufferedState;
-
-typedef enum {
-  META_WAYLAND_SURFACE_TOPLEVEL = 0,
-  META_WAYLAND_SURFACE_MAXIMIZED,
-  META_WAYLAND_SURFACE_FULLSCREEN,
-  META_WAYLAND_SURFACE_POPUP,
-} MetaWaylandSurfaceType;
-
-typedef struct
-{
-  MetaWaylandSurfaceType initial_type;
-  struct wl_resource *transient_for;
-  int x, y;
 
+  /* xdg_surface */
   char *title;
-  char *wm_class;
+  char *app_id;
 
+  /* gtk_surface */
   char *gtk_application_id;
   char *gtk_unique_bus_name;
   char *gtk_app_menu_path;
   char *gtk_menubar_path;
   char *gtk_application_object_path;
   char *gtk_window_object_path;
-} MetaWaylandSurfaceInitialState;
+} MetaWaylandDoubleBufferedState;
 
 typedef struct
 {
@@ -103,15 +91,11 @@ struct _MetaWaylandSurface
   MetaWaylandCompositor *compositor;
   MetaWaylandBufferReference buffer_ref;
   MetaWindow *window;
-  MetaWaylandSurfaceExtension *shell_surface;
+  MetaWaylandSurfaceExtension *xdg_surface;
   MetaWaylandSurfaceExtension *gtk_surface;
 
   /* All the pending state, that wl_surface.commit will apply. */
   MetaWaylandDoubleBufferedState pending;
-
-  /* All the initial state, that wl_shell_surface.set_* will apply
-     (through meta_window_new_for_wayland) */
-  MetaWaylandSurfaceInitialState *initial_state;
 };
 
 void                meta_wayland_init_shell     (MetaWaylandCompositor *compositor);
@@ -122,12 +106,12 @@ MetaWaylandSurface *meta_wayland_surface_create (MetaWaylandCompositor *composit
 						 guint32                version);
 void                meta_wayland_surface_free   (MetaWaylandSurface    *surface);
 
-void                meta_wayland_surface_set_initial_state (MetaWaylandSurface *surface,
-							    MetaWindow         *window);
-
 void                meta_wayland_surface_configure_notify (MetaWaylandSurface *surface,
 							   int                 width,
 							   int                 height,
 							   int                 edges);
 
+void                meta_wayland_surface_focused_set (MetaWaylandSurface *surface);
+void                meta_wayland_surface_focused_unset (MetaWaylandSurface *surface);
+
 #endif
